"""
cd ~/Autolab && bundle exec rails s -p 8000 --binding=0.0.0.0

To remove my shitty image:
docker rmi tango_python_tue
"""
from zipfile import ZipFile
from os.path import basename
import os
import inspect
import shutil
from jinja2 import Environment, FileSystemLoader
import glob
from unitgrade.framework import Report
from unitgrade_private import docker_helpers
from importlib.machinery import SourceFileLoader

# COURSES_BASE = "/home/tuhe/Autolab/courses/AutoPopulated"

CURDIR = os.path.dirname(__file__)
TEMPLATE_BASE = CURDIR + "/lab_template"

def jj(source, dest, data):
    """
    Run Jinja2 on the source template and output it to the destination file name
    using variables in 'data'. """
    if os.path.exists(dest) and os.path.samefile(source, dest):
        raise Exception()
    dir, f = os.path.split(source)
    file_loader = FileSystemLoader(dir)
    env = Environment(loader=file_loader)
    output = env.get_template(f).render(data)
    with open(dest, 'w') as f:
        f.write(output)
    return output

# def docker_build_image(tag='tango_python_tue'):
#     os.system(f"cd {CURDIR + '/../../../docker_images'}/docker_tango_python && docker build --tag {tag} .")
#     pass

def jj_handout(source, dest, data):
    out = jj(source, dest, data)
    shutil.copy(dest, dest+"-handout")
    return out


def zipFilesInDir(dirName, zipFileName, filter):
   with ZipFile(zipFileName, 'w') as zipObj:
       # Iterate over all the files in directory
       for folderName, subfolders, filenames in os.walk(dirName):
           for filename in filenames:
               if filter(filename):
                   # create complete filepath of file in directory
                   filePath = os.path.join(folderName, filename)
                   # Add file to zip
                   zipObj.write(filePath, basename(filePath))

def paths2report(base_path, report_file):
    mod = ".".join(os.path.relpath(report_file[:-3], base_path).split(os.sep))
    foo = SourceFileLoader(mod, report_file).load_module()
    # return foo.Report1
    # spec = importlib.util.spec_from_file_location(mod, report_file)
    # foo = importlib.util.module_from_spec(spec)
    for name, obj in inspect.getmembers(foo):
        if inspect.isclass(obj):
            # Last condition could be # and issubclass(obj, Report): but this is not safe when there are two
            # versions of unitgrade installed (git clone and pip installed package). So use this.
            if obj.__module__ == foo.__name__ and Report.__name__ in [c.__name__ for c in obj.mro()]:
                report = getattr(foo, name)
                return report
    return None

def run_relative(file, base):
    relative = os.path.relpath(file, base)
    mod = os.path.normpath(relative)[:-3].split(os.sep)
    print(mod)
    for pyver in ["python", "python3", "python3.9"]:
        cmd = f"cd {base} && {pyver} -m {'.'.join(mod)}"
        code = os.system(cmd)
        if code == 0:
            return code
    raise Exception("Could not run the file", file, "in dir", base, "using code", cmd)

def new_deploy_assignment(base_name, INSTRUCTOR_BASE, INSTRUCTOR_GRADE_FILE, STUDENT_BASE,
                          STUDENT_GRADE_FILE = None, # Defaults to instructor grade file.
                      output_tar=None,
                      COURSES_BASE=None,
                      autograde_image_tag='tango_python_tue',
                      student_should_upload_token=True,
                    homework_file=None,
                          description=None):
    if STUDENT_GRADE_FILE is None:
        STUDENT_GRADE_FILE = INSTRUCTOR_GRADE_FILE
        run_relative(STUDENT_GRADE_FILE, INSTRUCTOR_BASE) # Generate token in the instructor-directory. The student directory will not work.
    else:
        run_relative(STUDENT_GRADE_FILE, STUDENT_BASE)

    """ Check we got correct paths. """
    assert os.path.isfile(INSTRUCTOR_GRADE_FILE)
    assert os.path.isfile(STUDENT_GRADE_FILE)
    assert os.path.isdir(INSTRUCTOR_BASE)
    assert os.path.isdir(STUDENT_BASE)
    """ 
    In case the students should not upload the token, we do something else. 
    But what? 
    """
    if COURSES_BASE == None:
        COURSES_BASE = os.getcwd() + "/../tmp"
        if not os.path.exists(COURSES_BASE):
            os.mkdir(COURSES_BASE)

    LAB_DEST = os.path.join(COURSES_BASE, base_name)
    if homework_file is not None:
        student_should_upload_token = False

    # STUDENT_HANDOUT_DIR = os.path.dirname(STUDENT_GRADE_FILE) #"/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/students/programs"
    # INSTRUCTOR_GRADE_FILE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor/programs/report5.py"
    # Make instructor token file.
    # Get the instructor result file.
    run_relative(INSTRUCTOR_GRADE_FILE, INSTRUCTOR_BASE)
    f = glob.glob(os.path.dirname(INSTRUCTOR_GRADE_FILE) + "/*.token")[0]
    from unitgrade_private import load_token
    res, _ = load_token(f)

    # Now we have the instructor token file. Let's get the student token file.
    total_ = res['total'][1]
    problems = []
    # <<<<<<< HEAD
    problems.append(dict(name='Unitgrade score', description='Score obtained by automatic grading', max_score=total_, optional='false'))
    problems.append(dict(name='Written feedback', description='Written (TA) feedback', max_score=0, optional='true'))
    # print(problems)
    sc = [('Total', res['total'][0])] + [(q['title'], q['obtained']) for k, q in res['details'].items()]
    ss = ", ".join([f'"{t}": {s}' for t, s in sc])
    scores = '{"scores": {' + ss + '}}'


    # =======
    #     problems.append(dict(name='Unitgrade score', description='Automatic score as computed using the _grade.py script', max_score=total_, optional='false'))
    #     print(problems)
    #     sc = [('Total', res['total'][0])] + [(q['title'], q['obtained']) for k, q in res['details'].items()]
    #     ss = ", ".join([f'"{t}": {s}' for t, s in sc])
    #     scores = '{"scores": {' + ss + '}}'
    #     print(scores)
    #     # Quickly make student .token file to upload:
    #     # os.system(f"cd {os.path.dirname(STUDENT_HANDOUT_DIR)} && python -m programs.{os.path.basename(INSTRUCTOR_GRADE_FILE)[:-3]}")
    #     # os.system(f"cd {STUDENT_HANDOUT_DIR} && python {os.path.basename(INSTRUCTOR_GRADE_FILE)}")
    #     # handin_filename = os.path.basename(STUDENT_TOKEN_FILE)
    #     run_relative(os.path.join(STUDENT_BASE, STUDENT_GRADE_FILE), STUDENT_BASE)
    #     # if student_should_upload_token:
    # >>>>>>> 0429c721315832077f7682929c6f3a40449d85fc
    STUDENT_TOKEN_FILE = glob.glob(os.path.dirname(STUDENT_GRADE_FILE) + "/*.token")[0]
    handin_filename = os.path.basename(STUDENT_TOKEN_FILE)
    for _ in range(3):
        handin_filename = handin_filename[:handin_filename.rfind("_")]
    handin_filename += ".token"
    if not student_should_upload_token:
        student_token_src_filename = os.path.basename(handin_filename)
        handin_filename = os.path.basename(homework_file)

    print("> Name of handin file", handin_filename)
    if os.path.exists(LAB_DEST):
        shutil.rmtree(LAB_DEST)
    os.mkdir(LAB_DEST)
    assert os.path.exists(TEMPLATE_BASE)
    # Make the handout directory.
    # Start in the src directory. You should make the handout files first.
    os.mkdir(LAB_DEST + "/src")
    INSTRUCTOR_REPORT_FILE = INSTRUCTOR_GRADE_FILE[:-9] + ".py"
    if description is None:
        description = f'Upload the file {homework_file}' if homework_file is not None else handin_filename
    print("Making data...")
    # /home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor/programs/report5.py"
    data = {
        'base_name': base_name,
        'display_name': paths2report(INSTRUCTOR_BASE, INSTRUCTOR_REPORT_FILE).title,
        'handin_filename': handin_filename,
        # 'student_token_file': STUDENT_TOKEN_FILE,
        'autograde_image': autograde_image_tag,
        'src_files_to_handout': ['driver_python.py', handin_filename, # 'student_sources.zip',
                                 os.path.basename(docker_helpers.__file__),
                                 os.path.basename(INSTRUCTOR_GRADE_FILE),
                                 student_token_src_filename],  # Remove tname later; it is the upload.
        'instructor_grade_file': os.path.basename(INSTRUCTOR_GRADE_FILE),
        'grade_file_relative_destination': os.path.relpath(INSTRUCTOR_GRADE_FILE, INSTRUCTOR_BASE),
        'problems': problems,
        'student_should_upload_token': student_should_upload_token,
        'homework_file': homework_file,
        'student_token_src_filename': student_token_src_filename,
        'description': description
    }
    print("> Running jinja2")
    # shutil.copyfile(TEMPLATE_BASE + "/hello.yml", f"{LAB_DEST}/{base_name}.yml")
    # Figure out which of these are really needed and which are not.
    jj_handout(TEMPLATE_BASE + "/src/README", LAB_DEST + "/src/README", data)
    jj_handout(TEMPLATE_BASE + "/src/driver_python.py", LAB_DEST + "/src/driver_python.py", data)
    jj_handout(TEMPLATE_BASE + "/src/Makefile", LAB_DEST + "/src/Makefile", data)
    jj_handout(TEMPLATE_BASE + "/src/driver.sh", LAB_DEST + "/src/driver.sh", data)

    jj(TEMPLATE_BASE + "/Makefile", LAB_DEST + "/Makefile", data)
    jj(TEMPLATE_BASE + "/autograde-Makefile", LAB_DEST + "/autograde-Makefile", data=data)
    jj(TEMPLATE_BASE + "/hello.yml", f"{LAB_DEST}/{base_name}.yml", data=data)
    jj(TEMPLATE_BASE + "/hello.rb", f"{LAB_DEST}/{base_name}.rb", data=data)

    # Copy the student grade file to remove.
    shutil.copyfile(INSTRUCTOR_GRADE_FILE, f"{LAB_DEST}/src/{os.path.basename(INSTRUCTOR_GRADE_FILE)}")
    shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/src/{handin_filename}")
    shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/src/{student_token_src_filename}")
    # shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/src/{student_token_src_filename}-handout")

    import pathlib
    print("> Making archive..")
    # zip_base_dir = pathlib.Path(os.path.relpath(STUDENT_GRADE_FILE, STUDENT_BASE)).parent
    zip_base_dir = pathlib.Path(os.path.relpath(INSTRUCTOR_GRADE_FILE, INSTRUCTOR_BASE)).parent
    # Alternatively: Unzip the sources directory and hand it out.
    # shutil.make_archive(LAB_DEST + '/src/student_sources', 'zip', root_dir=STUDENT_BASE, base_dir=str(zip_base_dir))
    print("We made it")

    shutil.copyfile(docker_helpers.__file__, f"{LAB_DEST}/src/{os.path.basename(docker_helpers.__file__)}")
    os.mkdir(LAB_DEST + "/handin")
    os.mkdir(LAB_DEST + "/test-autograder")  # Otherwise make clean will screw up.
    print(f"cd {LAB_DEST} && make && cd {CURDIR}")
    # cmd = f"cd {LAB_DEST} && make && cd {CURDIR}"

    os.system(f"cd {LAB_DEST} && make && cd {CURDIR}")
    # os.system(f"cd {LAB_DEST} && make handout")
    # from slider.latexutils import latexmk
    # import subprocess
    # s = subprocess.check_output(cmd, shell=True)
    print("Ran make command...")
    if output_tar is None:
        output_tar = os.getcwd() + "/" + base_name + ".tar"
    # Making the writeup-directory.
    os.mkdir(LAB_DEST + "/writeup")

    writeup_template = f"""
    <html><body>
    To hand in this assignment, upload the file <b>{handin_filename}</b>
    </body></html>    
    """
    writeup = Environment(loader=FileSystemLoader("./")).from_string(writeup_template).render(data)
    with open(LAB_DEST + "/writeup/writeup.html", 'w') as f:
        f.write(writeup)

    # shutil.make_archive(output_tar[:-4], 'tar', root_dir=COURSES_BASE, base_dir=base_name)
    print("Log in to autolab, go to 'install assessment', upload the tar file", output_tar)
    # Lets try an alternative creation procedure.
    if os.path.exists(f"{LAB_DEST}/{base_name}-handout"):
        shutil.rmtree(f"{LAB_DEST}/{base_name}-handout")
    if os.path.exists(f"{LAB_DEST}/{base_name}-autograde"):
        shutil.rmtree(f"{LAB_DEST}/{base_name}-autograde")
    shutil.copytree(STUDENT_BASE, f"{LAB_DEST}/{base_name}-handout")
    shutil.copytree(STUDENT_BASE, f"{LAB_DEST}/{base_name}-autograde")
    shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/{base_name}-autograde/{student_token_src_filename}")

    jj(TEMPLATE_BASE + "/src/driver_python.py", f"{LAB_DEST}/{base_name}-autograde/driver_python.py", data)
    jj(TEMPLATE_BASE + "/autograde-Makefile", f"{LAB_DEST}/{base_name}-autograde/autograde-Makefile", data=data)
    # shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/{base_name}-autograde/{student_token_src_filename}") # Why is this needed? TBH it is not needed, really.
    shutil.copyfile(INSTRUCTOR_GRADE_FILE, f"{LAB_DEST}/{base_name}-autograde/{os.path.basename(INSTRUCTOR_GRADE_FILE)}")

    autograde_makefile_template = f"""
all:
	tar xf autograde.tar
	cp {handin_filename} {base_name}-autograde
	(cd {base_name}-autograde; python3 driver_python.py)

clean:
	rm -rf *~ {base_name}-autograde  
""".strip()
    autograde_makefile = Environment(loader=FileSystemLoader("./")).from_string(autograde_makefile_template).render(data)
    with open(f"{LAB_DEST}/autograde-Makefile", 'w') as f:
        f.write(autograde_makefile)
    # Check if you need to make the autograder...

    # Make the autograder and the handouts...
    shutil.make_archive(f"{LAB_DEST}/autograde", 'tar', root_dir=f"{LAB_DEST}", base_dir=f"{base_name}-autograde")
    shutil.make_archive(f"{LAB_DEST}/{base_name}-handout", 'zip', root_dir=f"{LAB_DEST}", base_dir=f"{base_name}-handout")
    shutil.make_archive(output_tar[:-4], 'tar', root_dir=COURSES_BASE, base_dir=base_name)
    return output_tar

def deploy_assignment(base_name, INSTRUCTOR_BASE, INSTRUCTOR_GRADE_FILE, STUDENT_BASE, STUDENT_GRADE_FILE,
                      output_tar=None,
                      COURSES_BASE=None,
                      autograde_image_tag='tango_python_tue',
                      student_should_upload_token=True):

    """ Check we got correct paths. """
    assert os.path.isfile(INSTRUCTOR_GRADE_FILE)
    assert os.path.isfile(STUDENT_GRADE_FILE)
    assert os.path.isdir(INSTRUCTOR_BASE)
    assert os.path.isdir(STUDENT_BASE)
    """ 
    In case the students should not upload the token, we do something else. 
    But what? 
    """

    # deploy_directly = COURSES_BASE != None
    if COURSES_BASE == None:
        COURSES_BASE = os.getcwd() + "/tmp"
        if not os.path.exists(COURSES_BASE):
            os.mkdir(COURSES_BASE)

    LAB_DEST = os.path.join(COURSES_BASE, base_name)

    # STUDENT_HANDOUT_DIR = os.path.dirname(STUDENT_GRADE_FILE) #"/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/students/programs"
    # INSTRUCTOR_GRADE_FILE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor/programs/report5.py"
    # Make instructor token file.
    # Get the instructor result file.
    run_relative(INSTRUCTOR_GRADE_FILE, INSTRUCTOR_BASE)
    f = glob.glob(os.path.dirname(INSTRUCTOR_GRADE_FILE) + "/*.token")[0]
    from unitgrade_private import load_token
    res, _ = load_token(f)



    # Now we have the instructor token file. Let's get the student token file.
    total_ = res['total'][1]
    problems = []
    problems.append(dict(name='Unitgrade score', description='', max_score=total_, optional='false') )
    # for k, q in res['details'].items():
    #     problems.append(dict(name=q['title'], description='', max_score=q['possible'], optional='true'))
    # problems.append(dict(name="Autograding Total", description='The description (set in autolab.py)', max_score=total_, optional='false'))
    print(problems)
    sc = [('Total', res['total'][0])] + [(q['title'], q['obtained']) for k, q in res['details'].items()]
    ss = ", ".join( [f'"{t}": {s}' for t, s in sc] )
    scores = '{"scores": {' + ss + '}}'
    print(scores)
    # Quickly make student .token file to upload:
    # os.system(f"cd {os.path.dirname(STUDENT_HANDOUT_DIR)} && python -m programs.{os.path.basename(INSTRUCTOR_GRADE_FILE)[:-3]}")
    # os.system(f"cd {STUDENT_HANDOUT_DIR} && python {os.path.basename(INSTRUCTOR_GRADE_FILE)}")
    # handin_filename = os.path.basename(STUDENT_TOKEN_FILE)
    run_relative(STUDENT_GRADE_FILE, STUDENT_BASE)
    STUDENT_TOKEN_FILE = glob.glob(os.path.dirname(STUDENT_GRADE_FILE) + "/*.token")[0]
    handin_filename = os.path.basename( STUDENT_TOKEN_FILE)
    for _ in range(3):
        handin_filename = handin_filename[:handin_filename.rfind("_")]
    handin_filename += ".token"

    print("> Name of handin file", handin_filename)
    if os.path.exists(LAB_DEST):
        shutil.rmtree(LAB_DEST)
    os.mkdir(LAB_DEST)
    assert os.path.exists(TEMPLATE_BASE)
    # Make the handout directory.
    # Start in the src directory. You should make the handout files first.
    os.mkdir(LAB_DEST + "/src")
    INSTRUCTOR_REPORT_FILE = INSTRUCTOR_GRADE_FILE[:-9] + ".py"
    a = 234
    # /home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor/programs/report5.py"
    data = {
            'base_name': base_name,
            'display_name': paths2report(INSTRUCTOR_BASE, INSTRUCTOR_REPORT_FILE).title,
            'handin_filename': handin_filename,
            'autograde_image': autograde_image_tag,
            'src_files_to_handout': ['driver_python.py', 'student_sources.zip', handin_filename, os.path.basename(docker_helpers.__file__),
                                     os.path.basename(INSTRUCTOR_GRADE_FILE)], # Remove tname later; it is the upload.
            'instructor_grade_file': os.path.basename(INSTRUCTOR_GRADE_FILE),
            'grade_file_relative_destination': os.path.relpath(INSTRUCTOR_GRADE_FILE, INSTRUCTOR_BASE),
            'problems': problems,
            }

    # shutil.copyfile(TEMPLATE_BASE + "/hello.yml", f"{LAB_DEST}/{base_name}.yml")
    # Figure out which of these are really needed and which are not.
    jj_handout(TEMPLATE_BASE + "/src/README", LAB_DEST + "/src/README", data)
    jj_handout(TEMPLATE_BASE + "/src/driver_python.py", LAB_DEST + "/src/driver_python.py", data)
    jj_handout(TEMPLATE_BASE + "/src/Makefile", LAB_DEST + "/src/Makefile",data)
    jj_handout(TEMPLATE_BASE + "/src/driver.sh", LAB_DEST + "/src/driver.sh",data)

    jj(TEMPLATE_BASE + "/Makefile", LAB_DEST + "/Makefile", data)
    jj(TEMPLATE_BASE + "/autograde-Makefile", LAB_DEST + "/autograde-Makefile",data=data)
    jj(TEMPLATE_BASE + "/hello.yml", f"{LAB_DEST}/{base_name}.yml", data=data)
    jj(TEMPLATE_BASE + "/hello.rb", f"{LAB_DEST}/{base_name}.rb", data=data)

    # Copy the student grade file to remove.
    shutil.copyfile(INSTRUCTOR_GRADE_FILE, f"{LAB_DEST}/src/{os.path.basename(INSTRUCTOR_GRADE_FILE)}")
    shutil.copyfile(STUDENT_TOKEN_FILE, f"{LAB_DEST}/src/{handin_filename}")

    shutil.make_archive(LAB_DEST + '/src/student_sources', 'zip', root_dir=STUDENT_BASE, base_dir=base_name)
    shutil.copyfile(docker_helpers.__file__, f"{LAB_DEST}/src/{os.path.basename(docker_helpers.__file__)}")
    os.mkdir(LAB_DEST +"/handin")
    os.mkdir(LAB_DEST +"/test-autograder") # Otherwise make clean will screw up.
    os.system(f"cd {LAB_DEST} && make && cd {CURDIR}")


    if output_tar is None:
        output_tar = os.getcwd() + "/" + base_name  + ".tar"

    shutil.make_archive(output_tar[:-4], 'tar', root_dir=COURSES_BASE, base_dir=base_name)
    return output_tar



def format_autolab_json(data, indent=None):
    import json

    stages = []
    pres = {
        "_presentation": "semantic",
        "stages": [], # "Build", "Test", "Timing"],
    }
    totals = {}
    for n, qs in data['details'].items():
        # print(n)
        title = qs['title']
        rs = {}
        for item, val in qs['items'].items():
            # print(item, val)
            item_name = val.get('nice_title', item[1]) # Attempt to give it a nicer title.

            pass_ = val['status'] == 'pass'
            d = {'passed': pass_}
            if not pass_:
                # Unfortunately, html is escaped in template, so linebreaks do not work.
                d['hint'] = val['stderr']
            rs[item_name] = d
        totals[title] = qs['obtained']
        stages.append(title)
        pres[title] = rs
    summary_key = "Summary"
    stages.append(summary_key)
    pres['stages'] = stages
    pres[summary_key] = totals
    # rs = {
    #     "_presentation": "semantic",
    #     "stages": ["Build", "Test", "Timing"],
    #     "Test": {
    #         "Add Things": {
    #             "passed": True
    #         },
    #         "Return Values": {
    #             "passed": False,
    #             "hint": "You need to return 1"
    #         }
    #     },
    #     'scores': 234,
    #       'pass': True
    # }
    if indent is not None: # for debug.
        json_out = json.dumps(pres, indent=2)
    else:
        json_out = json.dumps(pres)
    print(json_out)
    scores = {"scores": {'Unitgrade score': data['total'][0] }} #, 'scoreboard': [data['total'][0]] }
    print( json.dumps(scores) )

    a = 234
    pass



if __name__ == "__main__":
    """ For internal testing only. don't use this. """
    # print("Deploying to", COURSES_BASE)
    # docker_build_image()

    INSTRUCTOR_GRADE_FILE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor/programs/report1_grade.py"
    INSTRUCTOR_BASE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/instructor"

    STUDENT_BASE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/students"
    STUDENT_GRADE_FILE = "/home/tuhe/Documents/unitgrade_private_v1/examples/example_simplest/students/programs/report1_grade.py"

    output_tar = deploy_assignment("hello4", INSTRUCTOR_BASE, INSTRUCTOR_GRADE_FILE, STUDENT_BASE, STUDENT_GRADE_FILE=STUDENT_GRADE_FILE)



import os
import zipfile
import io
import numpy as np
import pickle

def unpack_sources_from_token(token_file, destination=None):
    from unitgrade_private import load_token

    rs, _ = load_token(token_file)
    if destination is None:
        destination = os.path.dirname(token_file)
    destination = os.path.normpath(destination)

    for k, data in rs['sources'].items():
        out = destination + "/" + os.path.basename(token_file)[:-6] + f"_{k}/"
        out = destination
        # if not os.path.exists(out):
        zf = zipfile.ZipFile(io.BytesIO(data['zipfile']))
        zf.extractall(out)

# Another token-blocking function


def get_coverage_files(token_file, instructor_grade_script_dir):
    """
    Get the files in the token which are part of the coverage analysis. This is useful for determining which files should be
    e.g. plagiarism checked.

    :param token_file:
    :param instructor_grade_script_dir:
    :return:
    """
    stoken, _ = load_token(token_file)
    cov_files = {}

    for q in stoken['details']:
        cov_files[q] = {}

        with open(f"{instructor_grade_script_dir}/unitgrade_data/{stoken['details'][q]['name']}.pkl", 'rb') as f:
            pk = pickle.load(f)
        for item in stoken['details'][q]['items']:
            key = (item, 'coverage')
            if key in pk:
                cov_files[q][key] = list( pk[(item, 'coverage')].keys() )
    return cov_files



def token_gather_hidden(token_rs, public_test_items_weight=1.):
    """
    Token-rs is a loaded token load_token(...)
    Return a version of the rs structure with the tokens blocked together.

    :param token_rs:
    :return:
    """

    r = token_rs
    rb = {'details': {}}
    ntot = 0
    for q in r['details']:
        rb['details'][q] = r['details'][q].copy()
        rb['details'][q]['items'] = {}
        for item in r['details'][q]['items']:
            if "hide" not in item[1] and 'hidden' not in item[1]:
                rb['details'][q]['items'][item] = [r['details'][q]['items'][item]]
                for i in r['details'][q]['items']:
                    if ("hide" in i[1] or 'hidden' in i[1]) and item[0] == i[0] and i[1].startswith(item[1]):
                        rb['details'][q]['items'][item].append(r['details'][q]['items'][i])
        # Now fix the score.
        item_scores = 0


        for item in rb['details'][q]['items']:
            # w0 = 1 if len(rb['details'][q]['items'][item]) == 1 else public_test_items_weight

            passed =  [i['status'] == 'pass' for i in rb['details'][q]['items'][item]  ]
            if len(passed) == 1:
                dt = np.mean(passed)
            else:
                dt = sum([(public_test_items_weight if k == 0 else 1) * e for k, e in enumerate(passed) ]) / (len(passed)-1+public_test_items_weight if len(passed) > 0 else 1)
            item_scores += dt

            if public_test_items_weight == 1:
                assert dt == np.mean( passed )

            # item_scores +=
        item_scores = item_scores / len(rb['details'][q]['items'])

        # This much score we got from the items.
        # print(q, ntot, item_scores)
        rb['details'][q]['obtained'] = item_scores * rb['details'][q]['possible']
        ntot += rb['details'][q]['obtained']
    rb['total'] = (ntot, r['total'][1])
    return rb

from unitgrade.utils import load_token

def determine_token_difference(student_token_rs, instructor_token_rs):
    # rsa, _ = load_token(student_token)
    # rsb, _ = load_token(instructor_token)
    rsa = student_token_rs
    rsb = instructor_token_rs
    # Must have same keys.
    assert rsa.keys() == rsb.keys()

    def where_a_better_b(a,b):
        a_better_than_b = {}
        kk = list(a.keys())
        kk +=  [k for k in b.keys() if b not in kk]
        for q in kk:
            for item in a[q]['items']:
                # print(q)
                if a[q]['items'][item]['status'] == 'pass' and (item not in b[q]['items'] or b[q]['items'][item]['status'] != 'pass'):
                    if q not in a_better_than_b:
                        a_better_than_b[q] = {'items': {}}
                    a_better_than_b[q]['items'][item] = {'a': a[q]['items'][item], 'b': b[q]['items'].get(item,None)}
        return a_better_than_b
    a_better_than_b = where_a_better_b(rsa['details'], rsb['details'])
    b_better_than_a = where_a_better_b(rsb['details'], rsa['details'])
    return a_better_than_b, b_better_than_a


def combine_token_results(token_a_rs, token_b_rs):
    """
    token_a_rs = load_token(...)
    token_b_rs = load_token(...)

    Combine by or'in the inputs. It will also recompute the token scores.


    :param token_a_rs:
    :param token_b_rs:
    :return:
    """
    rsd = {}
    n_tot = 0
    n_obt = 0

    for q in set(token_a_rs['details'].keys()) | set(token_b_rs['details'].keys()):
        itemsa = list(token_a_rs['details'][q]['items'])
        itemsb = list(token_b_rs['details'][q]['items'])
        items = itemsa + [i for i in itemsb if i not in itemsa]
        rsd[q] = {'items': {}}
        eql = True
        for i in items:
            ia = token_a_rs['details'][q]['items'].get(i, None) #[i]
            ib = token_b_rs['details'][q]['items'].get(i, None) #[i]
            if ia is None and ib is not None:
                item = ib
                eql = False
            elif ia is not None and ib is None:
                item = ia
                eql = False
            else:
                item = ia if ia['status'] == 'pass' else ib
                if ia['status'] != ib['status']:
                    eql = False
            rsd[q]['items'][i] = item


        for k in token_a_rs['details'][q].keys():
            if k not in ['obtained', 'items']:
                rsd[q][k] = token_a_rs['details'][q][k]
                assert token_a_rs['details'][q][k]  == token_b_rs['details'][q][k], k
            # rsd[q] = k

        w = token_a_rs['details'][q]['w']
        nc = int( np.floor( np.mean( [i['status'] == 'pass' for i in token_a_rs['details'][q]['items'].values()] ) * w ) )
        if eql:
            abt = token_a_rs['details'][q]['obtained']
            assert nc == token_a_rs['details'][q]['obtained'] and nc == token_b_rs['details'][q]['obtained'], f"points differ. {nc} != {abt}"
        rsd[q]['obtained'] = nc
        n_tot += w
        n_obt += nc


    rs = dict(total=(n_obt, n_tot), details=rsd, sources=None)
    return rs






# -*- coding: utf-8 -*-
#
#       Copyright 2013 Liftoff Software Corporation
#
# For license information see LICENSE.txt



# Meta
__version__ = '2.2'
__version_info__ = (2, 2)
__license__ = "GPLv3" # See LICENSE.txt
__author__ = 'Dan McDougall <daniel.mcdougall@liftoffsoftware.com>'

# TODO: Add the ability to mark variables, functions, classes, and methods for non-obfuscation.
# TODO: Add the ability to selectively obfuscate identifiers inside strings (for metaprogramming stuff).
# TODO: Add the ability to use a config file instead of just command line args.
# TODO: Add the ability to save a file that allows for de-obfuscation later (or at least the ability to debug).
# TODO: Separate out the individual functions of minification so that they can be chosen selectively like the obfuscation functions.

__doc__ = """\
**Python Minifier:**  Reduces the size of (minifies) Python code for use on
embedded platforms.

Performs the following:

    * Removes docstrings.
    * Removes comments.
    * Minimizes code indentation.
    * Removes trailing commas.
    * Joins multiline pairs of parentheses, braces, and brackets (and removes extraneous whitespace within).
    * Joins disjointed strings like, ``("some" "disjointed" "string")`` into single strings: ``('''some disjointed string''')
    * Preserves shebangs and encoding info (e.g. "# -- coding: utf-8 --").
    * Optionally, produces a bzip2 or gzip-compressed self-extracting python script containing the minified source for ultimate minification. *Added in version 1.4*
    * Optionally, obfuscates the code using the shortest possible combination of letters and numbers for one or all of class names, function/method names, and variables. The options are ``--obfuscate`` or ``-O`` to obfuscate everything, ``--obfuscate-variables``, ``--obfuscate-functions``, and ``--obfuscate-classes`` to obfuscate things individually (say, if you wanted to keep your module usable by external programs).  *Added in version 2.0*
    * Optionally, a value may be specified via --replacement-length to set the minimum length of random strings that are used to replace identifier names when obfuscating.
    * Optionally, if using Python 3, you may specify ``--nonlatin`` to use funky unicode characters when obfuscating. WARNING: This will result in some seriously hard-to-read code! **Tip:** Combine this setting with higher ``--replacement-length`` values to make the output even wackier.  *Added in version 2.0*
    * Pyminifier can now minify/obfuscate an arbitrary number of Python scripts in one go.  For example, ``pyminifier -O *.py`` will minify and obfuscate all files in the current directory ending in .py.  To prevent issues with using differentiated obfuscated identifiers across multiple files, pyminifier will keep track of what replaces what via a lookup table to ensure foo_module.whatever is gets the same replacement across all source files.  *Added in version 2.0*
    * Optionally, creates an executable zip archive (pyz) containing the minified/obfuscated source script and all implicit (local path) imported modules.  This mechanism automatically figures out which source files to include in the .pyz archive by analyzing the script passed to pyminifier on the command line (listing all the modules your script uses is unnecessary).  This is also the **ultimate** in minification/compression besting both the gzip and bzip2 compression mechanisms with the disadvantage that .pyz files cannot be imported into other Python scripts.  *Added in version 2.0*

Just how much space can be saved by pyminifier?  Here's a comparison:

    * The pyminifier source (all six files) takes up about 164k.
    * Performing basic minification on all pyminifier source files reduces that to ~104k.
    * Minification plus obfuscation provides a further reduction to 92k.
    * Minification plus the base64-encoded gzip trick (--gzip) reduces it to 76k.
    * Minification plus gzip compression plus obfuscation is also 76k (demonstrating that obfuscation makes no difference when compression algorthms are used).
    * Using the --pyz option on pyminifier.py creates a ~14k .pyz file that includes all the aforementioned files.

Various examples and edge cases are sprinkled throughout the pyminifier code so
that it can be tested by minifying itself.  The way to test is thus:

.. code-block:: bash

    $ python __main__.py __main__.py > minified_pyminifier.py
    $ python minified_pyminifier.py __main__.py > this_should_be_identical.py
    $ diff minified_pyminifier.py this_should_be_identical.py
    $

If you get an error executing minified_pyminifier.py or
``this_should_be_identical.py`` isn't identical to minified_pyminifier.py then
something is broken.

.. note::

    The test functions below are meaningless.  They only serve as test/edge
    cases for testing pyminifier.
"""

# Import built-in modules
import os, sys, re, io
from optparse import OptionParser
# from collections import Iterable

# Import our own modules
from . import minification
from . import token_utils
from . import obfuscate
from . import compression

py3 = False
lzma = False
if not isinstance(sys.version_info, tuple):
    if sys.version_info.major == 3:
        py3 = True
        try:
            import lzma
        except ImportError:
            pass

# Regexes
multiline_indicator = re.compile('\\\\(\s*#.*)?\n')

# The test.+() functions below are for testing pyminifier...
def test_decorator(f):
    """Decorator that does nothing"""
    return f

def test_reduce_operators():
    """Test the case where an operator such as an open paren starts a line"""
    (a, b) = 1, 2 # The indentation level should be preserved
    pass

def test_empty_functions():
    """
    This is a test function.
    This should be replaced with 'def test_empty_functions(): pass'
    """

class test_class(object):
    "Testing indented decorators"

    @test_decorator
    def test_function(self):
        pass

def test_function():
    """
    This function encapsulates the edge cases to prevent them from invading the
    global namespace.
    """
    # This tests method obfuscation:
    method_obfuscate = test_class()
    method_obfuscate.test_function()
    foo = ("The # character in this string should " # This comment
           "not result in a syntax error") # ...and this one should go away
    test_multi_line_list = [
        'item1',
        'item2',
        'item3'
    ]
    test_multi_line_dict = {
        'item1': 1,
        'item2': 2,
        'item3': 3
    }
    # It may seem strange but the code below tests our docstring removal code.
    test_string_inside_operators = imaginary_function(
        "This string was indented but the tokenizer won't see it that way."
    ) # To understand how this could mess up docstring removal code see the
    # minification.minification.remove_comments_and_docstrings() function
    # starting at this line:
    #     "elif token_type == tokenize.STRING:"
    # This tests remove_extraneous_spaces():
    this_line_has_leading_indentation    = '''<--That extraneous space should be
                                              removed''' # But not these spaces

def is_iterable(obj):
    """
    Returns `True` if *obj* is iterable but *not* if *obj* is a string, bytes,
    or a bytearray.
    """
    if isinstance(obj, (str, bytes, bytearray)):
        return False
    return isinstance(obj, Iterable)

def pyminify(options, files):
    """
    Given an *options* object (from `optparse.OptionParser` or similar),
    performs minification and/or obfuscation on the given *files* (any iterable
    containing file paths) based on said *options*.

    All accepted options can　be listed by running ``python __main__.py -h`` or
    examining the :py:func:`__init__.main` function.
    """
    global name_generator
    if not hasattr(files, '__iter__'): # is_iterable(files):
        print(
            "Error: The 'files' argument must be a list, tuple, etc of files.  "
            "Strings and bytes won't work.")
        sys.exit(1)
    if options.pyz:
        # Check to make sure we were only passed one script (only one at a time)
        if len(files) > 1:
            print("ERROR: The --pyz option only works with one python file at "
                  "a time.")
            print("(Dependencies will be automagically included in the "
                  "resulting .pyz)")
            sys.exit(1)
        # Make our .pyz:
        compression.zip_pack(files, options)
        return None # Make sure we don't do anything else
    # Read in our prepend text (if any)
    prepend = None
    if options.prepend:
        try:
            prepend = open(options.prepend).read()
        except Exception as err:
            print("Error reading %s:" % options.prepend)
            print(err)

    obfuscations = (options.obfuscate, options.obf_classes,
                    options.obf_functions, options.obf_variables,
                    options.obf_builtins, options.obf_import_methods)

    # Automatically enable obfuscation if --nonlatin (implied if no explicit
    # obfuscation is stated)
    if options.use_nonlatin and not any(obfuscations):
        options.obfuscate = True
    if len(files) > 1: # We're dealing with more than one file
        name_generator = None # So we can tell if we need to obfuscate
        if any(obfuscations):
            # Put together that will be used for all obfuscation functions:
            identifier_length = int(options.replacement_length)
            if options.use_nonlatin:
                if sys.version_info[0] == 3:
                    name_generator = obfuscate.obfuscation_machine(
                        use_unicode=True, identifier_length=identifier_length
                    )
                else:
                    print(
                        "ERROR: You can't use nonlatin characters without Python 3")
                    sys.exit(2)
            else:
                name_generator = obfuscate.obfuscation_machine(
                    identifier_length=identifier_length)
            table =[{}]
        cumulative_size = 0 # For size reduction stats
        cumulative_new = 0 # Ditto
        for sourcefile in files:
            # Record how big the file is so we can compare afterwards
            filesize = os.path.getsize(sourcefile)
            cumulative_size += filesize
            # Get the module name from the path
            module = os.path.split(sourcefile)[1]
            module = ".".join(module.split('.')[:-1])
            source = open(sourcefile).read()
            tokens = token_utils.listified_tokenizer(source)
            if not options.nominify: # Perform minification
                source = minification.minify(tokens, options)
            # Have to re-tokenize for obfucation (it is quick):
            tokens = token_utils.listified_tokenizer(source)
            # Perform obfuscation if any of the related options were set
            if name_generator:
                obfuscate.obfuscate(
                    module,
                    tokens,
                    options,
                    name_generator=name_generator,
                    table=table
                )
            # Convert back to text
            result = ''
            if prepend:
                result += prepend
            result += token_utils.untokenize(tokens)
            # Compress it if we were asked to do so
            if options.bzip2:
                result = compression.bz2_pack(result)
            elif options.gzip:
                result = compression.gz_pack(result)
            elif lzma and options.lzma:
                result = compression.lzma_pack(result)
            result += (
                "# Created by pyminifier "
                "(https://github.com/liftoff/pyminifier)\n")
            # Either save the result to the output file or print it to stdout
            if not os.path.exists(options.destdir):
                os.mkdir(options.destdir)
            # Need the path where the script lives for the next steps:
            filepath = os.path.split(sourcefile)[1]
            path = options.destdir + '/' + filepath # Put everything in destdir
            f = open(path, 'w')
            f.write(result)
            f.close()
            new_filesize = os.path.getsize(path)
            cumulative_new += new_filesize
            percent_saved = round((float(new_filesize) / float(filesize)) * 100, 2) if float(filesize)!=0 else 0
            print((
                "{sourcefile} ({filesize}) reduced to {new_filesize} bytes "
                "({percent_saved}% of original size)").format(**locals()))
        p_saved = round(
            (float(cumulative_new) / float(cumulative_size) * 100), 2)
        print("Overall size reduction: {0}% of original size".format(p_saved))
    else:
        # Get the module name from the path
        _file = files[0]
        module = os.path.split(_file)[1]
        module = ".".join(module.split('.')[:-1])
        filesize = os.path.getsize(_file)
        source = open(_file).read()
        # Convert the tokens from a tuple of tuples to a list of lists so we can
        # update in-place.
        tokens = token_utils.listified_tokenizer(source)
        if not options.nominify: # Perform minification
            source = minification.minify(tokens, options)
            # Convert back to tokens in case we're obfuscating
            tokens = token_utils.listified_tokenizer(source)
        # Perform obfuscation if any of the related options were set
        if options.obfuscate or options.obf_classes or options.obf_functions \
                or options.obf_variables or options.obf_builtins \
                or options.obf_import_methods:
            identifier_length = int(options.replacement_length)
            name_generator = obfuscate.obfuscation_machine(
                identifier_length=identifier_length)
            obfuscate.obfuscate(module, tokens, options)
        # Convert back to text
        result = ''
        if prepend:
            result += prepend
        result += token_utils.untokenize(tokens)
        # Compress it if we were asked to do so
        if options.bzip2:
            result = compression.bz2_pack(result)
        elif options.gzip:
            result = compression.gz_pack(result)
        elif lzma and options.lzma:
            result = compression.lzma_pack(result)
        result += (
            "# Created by pyminifier "
            "(https://github.com/liftoff/pyminifier)\n")
        # Either save the result to the output file or print it to stdout
        if options.outfile:
            f = io.open(options.outfile, 'w', encoding='utf-8')
            f.write(result)
            f.close()
            new_filesize = os.path.getsize(options.outfile)
            percent_saved = round(float(new_filesize)/float(filesize) * 100, 2)
            print((
                "{_file} ({filesize}) reduced to {new_filesize} bytes "
                "({percent_saved}% of original size)".format(**locals())))
        else:
            print(result)



import os
from unitgrade_private.version import __version__
from unitgrade_private.hidden_gather_upload import load_token, save_token
# from unitgrade_private.deployment import remove_hidden_methods
from unitgrade_private.plagiarism.mossit import unpack_sources_from_token
from unitgrade_private.hidden_create_files import setup_grade_file_report

# def cache_write(object, file_name, verbose=True):
#     assert False
#     dn = os.path.dirname(file_name)
#     if not os.path.exists(dn):
#         os.mkdir(dn)
#     if verbose: print("Writing cache...", file_name)
#     with open(file_name, 'wb', ) as f:
#         compress_pickle.dump(object, f, compression="lzma")
#     if verbose: print("Done!")
#
#
# def cache_exists(file_name):
#     assert False
#     return os.path.exists(file_name)
#
#
# def cache_read(file_name):
#     assert False
#     if os.path.exists(file_name):
#         with open(file_name, 'rb') as f:
#             return compress_pickle.load(f, compression="lzma")
#     else:
#         return None

